unit AFFVCLFrm;

interface

uses
  System.SysUtils, System.Types, System.UITypes, System.Classes, System.Variants,
  FMX.Types, FMX.Controls, FMX.Forms, FMX.Graphics, FMX.Dialogs, FMX.StdCtrls,
  FMX.Controls.Presentation, FMX.ScrollBox, FMX.Memo, FMX.FFEncode, FMX.FFDecode,
  FMX.FFBaseComponent, FMX.FFLog, FMX.FFBasePlay, FMX.FXPlay, FMX.Memo.Types;

type
  TForm1 = class(TForm)
    FFLogger1: TFFLogger;
    FFDecoder1: TFFDecoder;
    FFEncoder1: TFFEncoder;
    Memo1: TMemo;
    Panel1: TPanel;
    Button1: TButton;
    Button2: TButton;
    Button3: TButton;
    ProgressBar1: TProgressBar;
    Label1: TLabel;
    FXPlayer1: TFXPlayer;
    ImageControl1: TImageControl;
    Button4: TButton;
    FFDecoder2: TFFDecoder;
    ImageControl2: TImageControl;
    Button5: TButton;
    procedure FormCreate(Sender: TObject);
    procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
    procedure Button1Click(Sender: TObject);
    procedure Button2Click(Sender: TObject);
    procedure Button3Click(Sender: TObject);
    procedure Button4Click(Sender: TObject);
    procedure Button5Click(Sender: TObject);
    procedure FFLogger1Log(Sender: TObject; AThreadID: Integer;
      ALogLevel: TLogLevel; const ALogMsg: string);
    procedure FFEncoder1Terminate(Sender: TObject;
      const ATerminateInfo: TTerminateInfo);
    procedure FFEncoder1Progress(Sender: TObject; AProgressInfo: PProgressInfo);
    procedure FXPlayer1OpenFailed(Sender: TObject);
    procedure FXPlayer1FileOpen(Sender: TObject; const ADuration: Int64;
      AFrameWidth, AFrameHeight: Integer; var AScreenWidth,
      AScreenHeight: Integer);
    procedure FXPlayer1ClearVideo(Sender: TObject);
    procedure FXPlayer1DisplayVideo(Sender: TObject; ABitmap: TBitmap;
      const AAspectRatio: Single; const APTS: Int64);
    procedure FXPlayer1Position(Sender: TObject; const APTS: Int64);
  private
    { Private declarations }
    FAVLibPath: string;
    FSampleFile: string;
    FURL: string;
    FDuration: Int64;
    FLastView: string;
    procedure AddLog(S: string; const AShowLog: Boolean = True);
    procedure SwitchView(V: string);
  public
    { Public declarations }
  end;

var
  Form1: TForm1;

implementation

{$R *.fmx}

uses
  System.IOUtils,
  FMX.DialogService;

const
  SCaption = 'Android Demo of FFVCL %s - Delphi FFmpeg VCL Components';
  SWebSite = 'http://www.DelphiFFmpeg.com';

procedure TForm1.FormCreate(Sender: TObject);
begin
{$IFDEF ANDROID}
  FAVLibPath := TPath.GetHomePath + '/LibAV';
  FSampleFile := TPath.GetHomePath + '/sample.mp4';
{$ELSE}
  // you should assign the right path
  FAVLibPath := TPath.GetFullPath(ExtractFilePath(ParamStr(0)) + '../../../Bin/LibAV');
  FSampleFile := TPath.GetFullPath(ExtractFilePath(ParamStr(0)) + '../../sample.mp4');
{$ENDIF}
  FURL := 'rtsp://foo.bar/foo/bar';

  ImageControl1.Align := TAlignLayout.Client;
  ImageControl1.Bitmap.SetSize(Round(ImageControl1.Width), Round(ImageControl1.Height));
  ImageControl1.Visible := False;
  ImageControl2.Align := TAlignLayout.Client;
  ImageControl2.Bitmap.SetSize(Round(ImageControl2.Width), Round(ImageControl2.Height));
  ImageControl2.Visible := False;
  FLastView := 'decoder';
  Memo1.Visible := True;
  Memo1.Align := TAlignLayout.Client;

  Memo1.Text := Format(SCaption, [FFLogger1.Version]) + sLineBreak + SWebSite + sLineBreak + sLineBreak;
(*
{$IFDEF ANDROID}
  FFLogger.LogFile := TPath.GetTempPath + '/log.txt';
{$ELSE}
  FFLogger.LogFile := ExtractFilePath(ParamStr(0)) + '../../../Bin/log.txt';
{$ENDIF}
  FFLogger.LogToFile := True;
  FFLogger.Synchronous := True;
//  FFLogger.ClearLogFile;
*)
end;

procedure TForm1.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
begin
  FFEncoder1.Stop();
  FXPlayer1.Stop();
end;

procedure TForm1.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  FFLogger1.OnLog := nil;
end;

procedure TForm1.SwitchView(V: string);
begin
  if ImageControl1.Visible then
  begin
    if V <> 'decoder' then
      FLastView := 'decoder';
  end
  else if ImageControl2.Visible then
  begin
    if V <> 'player' then
      FLastView := 'player';
  end
  else
  begin
    if V <> 'log' then
      FLastView := 'log';
  end;

  if V = 'log' then
  begin
    ImageControl1.Visible := False;
    ImageControl2.Visible := False;
    Memo1.Visible := True;
  end
  else if v = 'decoder' then
  begin
    ImageControl2.Visible := False;
    Memo1.Visible := False;
    ImageControl1.Visible := True;
  end
  else if v = 'player' then
  begin
    ImageControl1.Visible := False;
    Memo1.Visible := False;
    ImageControl2.Visible := True;
  end
  else
    SwitchView(FLastView);
end;

procedure TForm1.AddLog(S: string; const AShowLog: Boolean);
begin
  Memo1.Lines.Add(S);
  if AShowLog then
    SwitchView('log');
end;

procedure TForm1.FFEncoder1Progress(Sender: TObject;
  AProgressInfo: PProgressInfo);
begin
  with AProgressInfo^ do
  begin
    if TotalDuration > 0 then
      ProgressBar1.Value := Round(CurrentDuration * ProgressBar1.Max / TotalDuration);
    Label1.Text := Format('fps = %d, s = %d, t = %d', [FPS, CurrentSize, CurrentDuration div 1000000]);
  end;
end;

procedure TForm1.FFEncoder1Terminate(Sender: TObject;
  const ATerminateInfo: TTerminateInfo);
begin
  if ATerminateInfo.Finished then
    ProgressBar1.Value := 100;
  if ATerminateInfo.TaskIndex < 0 then
    AddLog('Task done.');

  if ATerminateInfo.Exception then
    AddLog(ATerminateInfo.ExceptionMsg);
end;

procedure TForm1.FXPlayer1FileOpen(Sender: TObject; const ADuration: Int64;
  AFrameWidth, AFrameHeight: Integer; var AScreenWidth, AScreenHeight: Integer);
begin
  FDuration := ADuration;
  SwitchView('player');
end;

procedure TForm1.FXPlayer1OpenFailed(Sender: TObject);
begin
  AddLog(FXPlayer1.LastErrMsg);
end;

procedure TForm1.FXPlayer1Position(Sender: TObject; const APTS: Int64);
begin
  if FDuration > 0 then
    Label1.Text := Format('%f/%f', [APTS / 1000000, FDuration / 1000000])
  else
    Label1.Text := Format('%f', [APTS / 1000000])
end;

procedure ClearBitmap(ABitmap: TBitmap; AColor: TAlphaColor);
begin
  with ABitmap do
    if Canvas.BeginScene then
      try
        Canvas.Fill.Color := AColor;
        Canvas.FillRect(RectF(0, 0, Width, Height), 0, 0, [], 1);
      finally
        Canvas.EndScene;
{$IFDEF VER230}
        BitmapChanged;
{$ENDIF}
      end;
end;

procedure PaintToImage(AImage: TImageControl; ABitmap: TBitmap; AAspectRatio: Single);
var
  aspect_ratio: Single;
  h, w, x, y: Single;
begin
  // case keeping original
  aspect_ratio := AAspectRatio;
  // case custom, for example: 4/3, 16/9, 1.85, 2.35
//  aspect_ratio := 16/9;
  // case scaling to fit screen
//  aspect_ratio := AImage.Width / AImage.Height;

  // Image resized
  with AImage do
    if (Width <> Bitmap.Width) or (Height <> Bitmap.Width) then
    begin
      Bitmap.SetSize(Round(Width), Round(Height));
      ClearBitmap(Bitmap, TAlphaColorRec.Blue);
    end;
  // paint video to Image
  with AImage.Bitmap do
    if (Width > 0) and (Height > 0) then
    begin
      h := Height;
      w := Round(h * aspect_ratio) and (not 1);
      if w > Width then
      begin
        w := Width;
        h := Round(w / aspect_ratio) and (not 1);
      end;
      x := (Width - w) / 2;
      y := (Height - h) / 2;
      if Canvas.BeginScene then
        try
          Canvas.DrawBitmap(ABitmap,
            RectF(0, 0, ABitmap.Width, ABitmap.Height),
            RectF(x, y, x + w, y + h), 1);
        finally
          Canvas.EndScene;
{$IFDEF VER230}
          BitmapChanged;
{$ENDIF}
        end;
    end;
end;

procedure TForm1.FXPlayer1ClearVideo(Sender: TObject);
begin
  with ImageControl2 do
  begin
    if (Width <> Bitmap.Width) or (Height <> Bitmap.Width) then
      Bitmap.SetSize(Round(Width), Round(Height));
    ClearBitmap(Bitmap, TAlphaColorRec.Blue);
  end;
end;

procedure TForm1.FXPlayer1DisplayVideo(Sender: TObject; ABitmap: TBitmap;
  const AAspectRatio: Single; const APTS: Int64);
begin
  PaintToImage(ImageControl2, ABitmap, AAspectRatio);
end;

procedure TForm1.FFLogger1Log(Sender: TObject; AThreadID: Integer;
  ALogLevel: TLogLevel; const ALogMsg: string);
begin
  AddLog(ALogMsg, False);
end;

procedure TForm1.Button1Click(Sender: TObject);
var
  F: string;
  BMP: TBitmap;
begin
  F := FSampleFile;
  if F <> FFDecoder1.FileName then
    FFDecoder1.CloseFile;

  if not FFDecoder1.LoadAVLib(FAVLibPath) then
    AddLog(FFDecoder1.LastErrMsg)
  else if not TFile.Exists(F) then
    AddLog('File not exists: ' + F)
  else if not Assigned(FFDecoder1.FileHandle) or FFDecoder1.EOF then
  begin
    if not FFDecoder1.LoadFile(F) then
    begin
      AddLog(FFDecoder1.LastErrMsg);
      Exit;
    end;
    AddLog(FFDecoder1.FileInfoText, False);
    SwitchView('decoder');
  end;

  if not ImageControl1.Visible then
    SwitchView('decoder')
  else if not FFDecoder1.Decode() then
    AddLog(FFDecoder1.LastErrMsg)
  else
  begin
    BMP := TBitmap.Create;
    try
      if not FFDecoder1.CopyToBitmap(BMP) then
        AddLog(FFDecoder1.LastErrMsg)
      else
      begin
        PaintToImage(ImageControl1, BMP, BMP.Width / BMP.Height);
      end;
    finally
      BMP.Free;
    end;
  end;
end;

procedure TForm1.Button2Click(Sender: TObject);
var
  F: string;
  IO: TInputOptions;
  OO: TOutputOptions;
  LTaskIdx: Integer;
begin
  F := FSampleFile;
  if FFEncoder1.TasksCount > 0 then
  begin
    if FFEncoder1.Decoders[0, 0].FileName <> F then
      FFEncoder1.Stop(True)
    else if not FFEncoder1.Terminated then
    begin
      SwitchView('log');
      Exit;
    end;
  end;

  SwitchView('log');
  if not FFEncoder1.LoadAVLib(FAVLibPath) then
    AddLog(FFEncoder1.LastErrMsg)
  else if not TFile.Exists(F) then
    AddLog('File not exists: ' + F)
  else
  begin
    InitInputOptions(@IO);
    //IO.Options := '-re';
    FFEncoder1.ClearTasks;
    LTaskIdx := FFEncoder1.AddTask(F, @IO);
    if LTaskIdx < 0 then
      AddLog('Failed to AddTask: ' + FFEncoder1.LastErrMsg)
    else
    begin
      AddLog(FFEncoder1.Decoders[LTaskIdx, 0].FileInfoText, False);
      InitOutputOptions(@OO);
      OO.Options := '-c:a aac -c:v mpeg4';
      //OO.Options := '-c:a aac -c:v libx264'; // GPL required
      //FFEncoder1.PreviewVideo := True;
      if FFEncoder1.SetOutput(LTaskIdx, TPath.Combine(TPath.GetTempPath, ExtractFileName(F)), @OO) < 0 then
      begin
        AddLog('Failed to SetOutput: ' + FFEncoder1.LastErrMsg);
        FFEncoder1.RemoveTask(LTaskIdx);
      end
      else
      begin
        ProgressBar1.Value := 0;
        FFEncoder1.Start(1);
      end;
    end;
  end;
end;

procedure TForm1.Button3Click(Sender: TObject);
var
  F: string;
begin
  F := FSampleFile;

  if (FXPlayer1.FileName = F) and not (FXPlayer1.PlayState in [psStop, psEnd]) then
  begin
    if ImageControl2.Visible then
    begin
      if FXPlayer1.PlayState <> psPause then
        FXPlayer1.Pause
      else
        FXPlayer1.Resume;
    end
    else
      SwitchView('player');
    Exit;
  end;

  if not FXPlayer1.LoadAVLib(FAVLibPath) then
    AddLog(FXPlayer1.LastErrMsg)
  else if not TFile.Exists(F) then
    AddLog('File not exists: ' + F)
  else
    FXPlayer1.TryOpen(F);
end;

procedure TForm1.Button4Click(Sender: TObject);
begin
  if SameText(FXPlayer1.FileName, FURL) and not (FXPlayer1.PlayState in [psStop, psEnd]) then
  begin
    if ImageControl2.Visible then
      FXPlayer1.Stop()
    else
      SwitchView('player');
    Exit;
  end;

  if not FXPlayer1.LoadAVLib(FAVLibPath) then
    AddLog(FXPlayer1.LastErrMsg)
  else
  begin
    TDialogService.InputQuery('FXPlayer', ['URL: http, rtsp, rtmp, etc.'], [FURL],
      procedure(const AResult: TModalResult; const AValues: array of string)
      begin
        if AResult = mrOk then
        begin
          FURL := AValues[0];
          FXPlayer1.DefaultOptions('-fflags nobuffer -analyzeduration 10000 -max_delay 0 -fpsprobesize 5 -probesize 32');
          FXPlayer1.TryOpen(FURL);
        end;
      end
      );
  end;
end;

procedure TForm1.Button5Click(Sender: TObject);
begin
  if Memo1.Visible then
    SwitchView(FLastView)
  else
    SwitchView('log');
end;

end.
